import typing
from contextlib import asynccontextmanager, contextmanager

from context_handler import _datastructures, exc

T = typing.TypeVar("T")


class SyncContext(typing.Generic[T]):
    def __init__(self, provider: _datastructures.Provider[T]) -> None:
        self._provider = provider
        self._inside_ctx = False
        self._client: typing.Optional[T] = None

    def in_context(self):
        if self._client is None:
            return False
        return not self.get_provider().is_closed(self._client)

    @property
    def client(self) -> T:
        if self._client is None:
            raise exc.ContextNotInitializedError
        return self._client

    def _set_client(self, client: T):
        self._client = client

    def _reset_context(self):
        if self._client is None:
            return
        if not self.get_provider().is_closed(self._client):
            self.get_provider().close_client(self._client)
        self._set_client(None)
        self._inside_ctx = False

    def _open_context(self):
        error = None
        with self.get_provider().acquire() as client:
            self._set_client(client)
            self._inside_ctx = True
            try:
                yield
            except Exception as err:
                error = err
        self._reset_context()
        if error is not None:
            raise error

    def _begin_context(self):
        error = None
        with self.get_provider().acquire() as client:
            try:
                yield client
            except Exception as err:
                error = err
        if error is not None:
            raise error

    def _contexted_begin(self):
        yield self.client

    def _contexted_open(self):
        yield

    @contextmanager
    def begin(self):
        if self.in_context():
            return self._contexted_begin()
        return self._begin_context()

    @contextmanager
    def open(self):
        if self.in_context():
            return self._contexted_open()
        return self._open_context()

    def __getattribute__(self, name: str) -> typing.Any:
        if name == "_provider":
            name = "invalid"
        return super().__getattribute__(name)

    def get_provider(self) -> _datastructures.ImmutableSyncProvider[T]:
        return _datastructures.ImmutableSyncProvider(
            super().__getattribute__("_provider")
        )


class AsyncContext(typing.Generic[T]):
    def __init__(self, provider: _datastructures.AsyncProvider[T]) -> None:
        self._provider = provider
        self._inside_ctx = False
        self._client: typing.Optional[T] = None

    def in_context(self):
        if self._client is None:
            return False
        return not self.get_provider().is_closed(self._client)

    @property
    def client(self) -> T:
        if self._client is None:
            raise exc.ContextNotInitializedError
        return self._client

    def _set_client(self, client: T):
        self._client = client

    async def _reset_context(self):
        if self._client is None:
            return
        if not self.get_provider().is_closed(self._client):
            await self.get_provider().close_client(self._client)
        self._set_client(None)
        self._inside_ctx = False

    async def _open_context(self):
        error = None
        async with self.get_provider().acquire() as client:
            self._set_client(client)
            self._inside_ctx = True
            try:
                yield
            except Exception as err:
                error = err
        await self._reset_context()
        if error is not None:
            raise error

    async def _begin_context(self):
        error = None
        async with self.get_provider().acquire() as client:
            try:
                yield client
            except Exception as err:
                error = err
        if error is not None:
            raise error

    async def _contexted_begin(self):
        yield self.client

    async def _contexted_open(self):
        yield

    @asynccontextmanager
    def begin(self):
        if self.in_context():
            return self._contexted_begin()
        return self._begin_context()

    @asynccontextmanager
    def open(self):
        if self.in_context():
            return self._contexted_open()
        return self._open_context()

    def __getattribute__(self, name: str) -> typing.Any:
        if name == "_provider":
            name = "invalid"
        return super().__getattribute__(name)

    def get_provider(self) -> _datastructures.ImmutableAsyncProvider[T]:
        return _datastructures.ImmutableAsyncProvider(
            super().__getattribute__("_provider")
        )



import enum
import typing

T = typing.TypeVar("T")


@typing.runtime_checkable
class Provider(typing.Protocol[T]):
    """Client Adapter Interface Accepted by SyncContext"""

    state_name: str

    def is_closed(self, client: T) -> bool:
        """Returns if client is closed or released"""

    def close_client(self, client: T) -> None:
        """Closes/releases client"""

    def acquire(self) -> typing.ContextManager[T]:
        """Acquires a client `T` and releases at the end"""


@typing.runtime_checkable
class AsyncProvider(typing.Protocol[T]):
    """Client Adapter Interface Accepted by AsyncContext"""

    state_name: str

    def is_closed(self, client: T) -> bool:
        """Returns if client is closed or released"""

    async def close_client(self, client: T) -> None:
        """Closes/releases client"""

    def acquire(self) -> typing.AsyncContextManager[T]:
        """Acquires a client `T` and releases at the end"""


class ImmutableSyncProvider(typing.Generic[T]):
    """Wrapper class to prevent mutating provider"""

    state_name = "immutable_sync_provider"

    def __init__(self, provider: Provider[T]) -> None:
        self._provider = provider

    def __getattribute__(self, name: str) -> typing.Any:
        allowed = ["get_state_name", "is_closed", "close_client", "acquire"]
        if name in allowed:
            return super().__getattribute__(name)
        message = "'{}' object has no attribute '{}'"
        raise AttributeError(
            message.format(ImmutableSyncProvider.__name__, name)
        )

    def get_state_name(self):
        return super().__getattribute__("_provider").state_name

    def is_closed(self, client: T) -> bool:
        return super().__getattribute__("_provider").is_closed(client)

    def close_client(self, client: T) -> None:
        return super().__getattribute__("_provider").close_client(client)

    def acquire(self) -> typing.ContextManager[T]:
        return super().__getattribute__("_provider").acquire()


class ImmutableAsyncProvider(typing.Generic[T]):
    """Wrapper class to prevent mutating provider"""

    state_name = "immutable_async_provider"

    def __init__(self, provider: AsyncProvider[T]) -> None:
        self._provider = provider

    def __getattribute__(self, name: str) -> typing.Any:
        allowed = ["get_state_name", "is_closed", "close_client", "acquire"]
        if name in allowed:
            return super().__getattribute__(name)
        message = "'{}' object has no attribute '{}'"
        raise AttributeError(
            message.format(ImmutableSyncProvider.__name__, name)
        )

    def get_state_name(self):
        return super().__getattribute__("_provider").state_name

    def is_closed(self, client: T) -> bool:
        return super().__getattribute__("_provider").is_closed(client)

    def close_client(self, client: T) -> typing.Coroutine[None, None, None]:
        return super().__getattribute__("_provider").close_client(client)

    def acquire(self) -> typing.AsyncContextManager[T]:
        return super().__getattribute__("_provider").acquire()


@typing.runtime_checkable
class AbstractSyncContext(typing.Protocol[T]):
    _provider: Provider[T]
    _inside_ctx: bool

    def __init__(self, provider: Provider[T]) -> None:
        ...

    def in_context(self) -> bool:
        """Returns if `.open()` or `.begin()` calls where made inside an open context"""

    @property
    def client(self) -> T:
        """Returns a client instance if context is open"""

    def open(self) -> typing.ContextManager[None]:
        """Opens context"""

    def begin(self) -> typing.ContextManager[T]:
        """Returns client from open context or a independent client if no context is open."""

    def get_provider(self) -> ImmutableSyncProvider[T]:
        """Returns internal provider"""


@typing.runtime_checkable
class AbstractAsyncContext(typing.Protocol[T]):
    _provider: AsyncProvider[T]
    _inside_ctx: bool

    def __init__(self, provider: AsyncProvider[T]) -> None:
        ...

    def in_context(self) -> bool:
        """Returns if `.open()` or `.begin()` calls where made inside an open context"""

    @property
    def client(self) -> T:
        """Returns a client instance if context is open"""

    def open(self) -> typing.AsyncContextManager[None]:
        """Opens context"""

    def begin(self) -> typing.AsyncContextManager[T]:
        """Returns client from open context or a independent client if no context is open."""

    def get_provider(self) -> ImmutableAsyncProvider[T]:
        """Returns internal provider"""


AbstractContext = typing.TypeVar(
    "AbstractContext", AbstractSyncContext, AbstractAsyncContext
)


class _StateApp(typing.Protocol):
    state: typing.Type


class _ContextApp(typing.Protocol):
    context: typing.Type


class _CtxApp(typing.Protocol):
    ctx: typing.Type


class _HasState(typing.Protocol):
    state: typing.Type
    app: "_StateApp"


class _HasContext(typing.Protocol):
    context: typing.Type
    app: "_ContextApp"


class _HasCtx(typing.Protocol):
    ctx: typing.Type
    app: "_CtxApp"


StateApp = typing.Union[_StateApp, _ContextApp, _CtxApp]

HasState = typing.Union[_HasState, _HasContext, _HasCtx]


class StateWrapper:
    _valid_state_attrs = ["state", "context", "ctx"]

    def __init__(self, has_state: HasState) -> None:
        self.has_state = has_state
        self._validate_instance()
        self._instance_state_attr = self._get_state_attr(self.has_state)
        self._app_state_attr = self._get_state_attr(self.has_state.app)

    def _validate_instance(self):
        if not hasattr(self.has_state, "app"):
            raise TypeError("State Handler must have 'app' attribute")

    def _get_state_attr(
        self,
        instance: typing.Union[HasState, StateApp],
    ):
        for item in self._valid_state_attrs:
            if hasattr(instance, item):
                return item
        else:
            raise NotImplementedError(
                "State Handler does not have supported state_attrs"
            )

    @property
    def _app_state(self):
        return getattr(self.has_state.app, self._app_state_attr)

    @property
    def _instance_state(self):
        return getattr(self.has_state, self._app_state_attr)

    @staticmethod
    def _get(state: type, name: str, _cast: typing.Type[T]) -> typing.Optional[T]:
        return getattr(state, name, None)

    @staticmethod
    def _set(state: type, name: str, val: typing.Any):
        setattr(state, name, val)

    def app_get(
        self, name: str, _cast: typing.Type[T] = typing.Any
    ) -> typing.Optional[T]:
        return self._get(self._app_state, name, _cast)

    def get(
        self, name: str, _cast: typing.Type[T] = typing.Any
    ) -> typing.Optional[T]:
        return self._get(self._instance_state, name, _cast)

    def app_set(self, name: str, val: typing.Any):
        self._set(self._app_state, name, val)

    def set(self, name: str, val: typing.Any):
        self._set(self._instance_state, name, val)


class AbstractSyncContextFactory(typing.Protocol[T]):
    """Creates a Context Factory to handle contexts inside a state"""

    _provider_class: typing.Type[Provider[T]]
    _context_class: typing.Type[AbstractSyncContext[T]]
    _state_name: typing.Optional[str]

    def _get_context(
        self, state_wrapper: StateWrapper
    ) -> AbstractSyncContext[T]:
        """Initializes Context"""

    def generate_state_name(self) -> str:
        """Returns a key name to store context in state"""

    def has_active_context(
        self, has_state: HasState
    ) -> typing.Optional[AbstractSyncContext[T]]:
        """Returns context from `has_state` if context in has_state, else None"""

    def _set_active_context(
        self, context: AbstractSyncContext[T], state_wrapper: StateWrapper
    ):
        """Sets context in state"""

    def __call__(self, has_state: HasState) -> AbstractSyncContext[T]:
        """Returns context from has_state if exists or opens new context, stores in state, and then returns state"""

    def from_provider(self, provider: typing.Type[Provider[T]]):
        """Returns context from a given provider"""


class AbstractAsyncContextFactory(typing.Protocol[T]):
    """Creates a Context Factory to handle contexts inside a state"""

    _provider_class: typing.Type[AsyncProvider[T]]
    _context_class: typing.Type[AbstractAsyncContext[T]]
    _state_name: typing.Optional[str]

    def _get_context(
        self, state_wrapper: StateWrapper
    ) -> AbstractAsyncContext[T]:
        """Initializes Context"""

    def generate_state_name(self) -> str:
        """Returns a key name to store context in state"""

    def has_active_context(
        self, has_state: HasState
    ) -> typing.Optional[AbstractAsyncContext[T]]:
        """Returns context from `has_state` if context in has_state, else None"""

    def _set_active_context(
        self, context: AbstractAsyncContext[T], state_wrapper: StateWrapper
    ):
        """Sets context in state"""

    def __call__(self, has_state: HasState) -> AbstractAsyncContext[T]:
        """Returns context from has_state if exists or opens new context, stores in state, and then returns state"""

    def from_provider(self, provider: typing.Type[AsyncProvider[T]]):
        """Returns context from a given provider"""


class ContextGetter:
    class ArgType(enum.Enum):
        INSTANCE = enum.auto()
        CONTEXT = enum.auto()
        HAS_STATE = enum.auto()
        VIEW = HAS_STATE

        @classmethod
        def get(cls, name: str):
            return getattr(cls, name.upper())

    def __init__(
        self,
        arg_type: ArgType,
        *,
        context_attr_name: typing.Optional[str] = None,
        _factory: typing.Optional[
            typing.Union[
                AbstractAsyncContextFactory,
                AbstractSyncContextFactory,
            ]
        ] = None,
    ) -> None:
        self.arg_type = arg_type
        self.context_attr_name = context_attr_name
        self.factory = _factory

    def _instance(self, instance):
        return getattr(instance, self.context_attr_name)  # type: ignore

    def _context(self, context):
        return context

    def _has_state(self, has_state):
        return self.factory(has_state)  # type: ignore

    def get(self, first_arg):
        return getattr(self, "_{}".format(self.arg_type.name.lower()))(
            first_arg
        )



import typing
from functools import partial, wraps
from inspect import isasyncgenfunction, iscoroutinefunction
from types import SimpleNamespace

from typing_extensions import ParamSpec

from . import _datastructures, context

T = typing.TypeVar("T")
P = ParamSpec("P")
ClassT = typing.TypeVar("ClassT")


def _open_sync_ctx_in_sync(
    getter: _datastructures.ContextGetter,
    func,
    first_arg,
    *args,
    **kwargs,
):
    with getter.get(first_arg).open():
        return func(first_arg, *args, **kwargs)  # type: ignore


async def _open_sync_ctx_in_coro(
    getter: _datastructures.ContextGetter,
    func,
    first_arg,
    *args,
    **kwargs,
):
    with getter.get(first_arg).open():
        return await func(first_arg, *args, **kwargs)  # type: ignore


async def _open_sync_ctx_in_asyncgen(
    getter: _datastructures.ContextGetter,
    func,
    first_arg,
    *args,
    **kwargs,
):
    with getter.get(first_arg).open():
        async for item in func(first_arg, *args, **kwargs):  # type: ignore
            yield item


async def _open_async_ctx_in_coro(
    getter: _datastructures.ContextGetter,
    func,
    first_arg,
    *args,
    **kwargs,
):
    async with getter.get(first_arg).open():
        return await func(first_arg, *args, **kwargs)  # type: ignore


async def _open_async_ctx_in_async_gen(
    getter: _datastructures.ContextGetter,
    func,
    first_arg,
    *args,
    **kwargs,
):
    async with getter.get(first_arg).open():
        async for item in func(first_arg, *args, **kwargs):  # type: ignore
            yield item


def _setup_wrapper(wrapper: typing.Callable, func, context_getter):
    @wraps(func)
    def wrapped(*args, **kwargs):
        return wrapper(context_getter, func, *args, **kwargs)

    return wrapped


def _get_sync_wrapper(func):
    inner = _open_sync_ctx_in_sync
    if isasyncgenfunction(func):
        inner = _open_sync_ctx_in_asyncgen
    elif iscoroutinefunction(func):
        inner = _open_sync_ctx_in_coro
    return inner


def _get_async_wrapper(func):
    if iscoroutinefunction(func):
        inner = _open_async_ctx_in_coro
    elif isasyncgenfunction(func):
        inner = _open_async_ctx_in_async_gen
    else:
        raise TypeError("AsyncContext cannot be used in sync function")
    return inner


@typing.overload
def sync_context(
    *,
    first_arg_type: typing.Literal["instance"],
    context_attr_name: str,
) -> typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]:
    ...


@typing.overload
def sync_context(
    *,
    first_arg_type: typing.Literal["view"],
    _factory: typing.Union[
        _datastructures.AbstractAsyncContextFactory,
        _datastructures.AbstractSyncContextFactory,
    ],
) -> typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]:
    ...


@typing.overload
def sync_context(
    *,
    first_arg_type: typing.Literal["context"],
) -> typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]:
    ...


@typing.overload
def sync_context(
    func: typing.Callable[P, T],
    /,
) -> typing.Callable[P, T]:
    ...


def sync_context(
    func: typing.Callable[P, T] = None,
    /,
    *,
    first_arg_type: typing.Union[
        typing.Literal["instance"],
        typing.Literal["view"],
        typing.Literal["context"],
    ] = "context",
    **kwargs,
) -> typing.Union[
    typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]],
    typing.Callable[P, T],
]:
    context_getter = _datastructures.ContextGetter(
        _datastructures.ContextGetter.ArgType.get(first_arg_type),
        **kwargs,
    )

    def outer(func):
        return _setup_wrapper(_get_sync_wrapper(func), func, context_getter)

    if func:
        return outer(func)
    return outer


@typing.overload
def async_context(
    *,
    first_arg_type: typing.Literal["instance"],
    context_attr_name: str,
) -> typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]:
    ...


@typing.overload
def async_context(
    *,
    first_arg_type: typing.Literal["view"],
    _factory: typing.Union[
        _datastructures.AbstractAsyncContextFactory,
        _datastructures.AbstractSyncContextFactory,
    ],
) -> typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]:
    ...


@typing.overload
def async_context(
    *,
    first_arg_type: typing.Literal["context"],
) -> typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]:
    ...


@typing.overload
def async_context(
    func: typing.Callable[P, T],
    /,
) -> typing.Callable[P, T]:
    ...


def async_context(
    func: typing.Callable[P, T] = None,
    /,
    *,
    first_arg_type: typing.Union[
        typing.Literal["instance"],
        typing.Literal["view"],
        typing.Literal["context"],
    ] = "context",
    **kwargs,
) -> typing.Union[
    typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]],
    typing.Callable[P, T],
]:
    context_getter = _datastructures.ContextGetter(
        _datastructures.ContextGetter.ArgType.get(first_arg_type),
        **kwargs,
    )

    def outer(func):
        return _setup_wrapper(_get_async_wrapper(func), func, context_getter)

    if func:
        return outer(func)  # type: ignore
    return outer


def _guess_context_class(
    provider_class: typing.Type[
        typing.Union[_datastructures.Provider, _datastructures.AsyncProvider]
    ]
):
    if issubclass(provider_class, _datastructures.AsyncProvider):
        return context.AsyncContext
    elif issubclass(provider_class, _datastructures.Provider):
        return context.SyncContext
    raise TypeError(
        "provider_class must implement either _datastructures.Provider or _datastructures.AsyncProvider protocol"
    )


__all__ = ["sync_context", "async_context"]



import typing

from context_handler import _datastructures, exc

T = typing.TypeVar("T")


class _ContextFactory(typing.Generic[T]):
    def __init__(
        self,
        provider_class: typing.Union[
            typing.Type[_datastructures.AsyncProvider[T]],
            typing.Type[_datastructures.Provider[T]],
        ],
        context_class: typing.Union[
            typing.Type[_datastructures.AbstractAsyncContext[T]],
            typing.Type[_datastructures.AbstractSyncContext[T]],
        ],
        context_state_name: typing.Optional[str] = None,
    ) -> None:
        self._provider_class = provider_class
        self._context_class = context_class
        self._state_name = context_state_name or self.generate_state_name(
            self._provider_class
        )

    @staticmethod
    def generate_state_name(
        provider_class: typing.Union[
            typing.Type[_datastructures.AsyncProvider[T]],
            typing.Type[_datastructures.Provider[T]],
        ]
    ):
        return provider_class.state_name.lower().replace("provider", "context")

    def has_active_context(self, has_state: _datastructures.HasState):
        return bool(
            self._get_active_context(_datastructures.StateWrapper(has_state))
        )

    def _get_active_context(self, state_wrapper: _datastructures.StateWrapper):
        return state_wrapper.get(self._state_name, self._context_class)

    def _set_active_context(
        self,
        context: _datastructures.AbstractContext,
        state_wrapper: _datastructures.StateWrapper,
    ):
        state_wrapper.set(self._state_name, context)

    def _get_context(self, state_wrapper: _datastructures.StateWrapper):
        provider = state_wrapper.app_get(
            self._provider_class.state_name, self._provider_class
        )
        if not isinstance(
            provider, (_datastructures.AsyncProvider, _datastructures.Provider)
        ):
            raise exc.NoProviderInState(
                "State Handler does not have provider instantiated"
            )
        return self._context_class(provider)  # type: ignore

    def __call__(self, has_state: _datastructures.HasState):
        state_wrapper = _datastructures.StateWrapper(has_state)
        if (context := self._get_active_context(state_wrapper)) is not None:
            return context
        context = self._get_context(state_wrapper)
        self._set_active_context(context, state_wrapper)
        return context

    def from_provider(
        self,
        provider: typing.Union[
            typing.Type[_datastructures.AsyncProvider[T]],
            typing.Type[_datastructures.Provider[T]],
        ],
    ) -> typing.Union[
        _datastructures.AbstractAsyncContext[T],
        _datastructures.AbstractSyncContext[T],
    ]:
        return self._context_class(provider)  # type: ignore


@typing.overload
def context_factory(
    provider_class: typing.Type[_datastructures.AsyncProvider[T]],
    context_class: typing.Type[_datastructures.AbstractAsyncContext[T]],
    context_state_name: typing.Optional[str] = None,
) -> _datastructures.AbstractAsyncContextFactory[T][T]:
    ...


@typing.overload
def context_factory(
    provider_class: typing.Type[_datastructures.Provider[T]],
    context_class: typing.Type[_datastructures.AbstractSyncContext[T]],
    context_state_name: typing.Optional[str] = None,
) -> _datastructures.AbstractSyncContextFactory[T][T]:
    ...


@typing.overload
def context_factory(
    provider_class: typing.Union[
        typing.Type[_datastructures.AsyncProvider[T]],
        typing.Type[_datastructures.Provider[T]],
    ],
    context_class: typing.Union[
        typing.Type[_datastructures.AbstractAsyncContext[T]],
        typing.Type[_datastructures.AbstractSyncContext[T]],
    ],
    context_state_name: typing.Optional[str] = None,
) -> typing.Union[
    _datastructures.AbstractAsyncContextFactory[T],
    _datastructures.AbstractSyncContextFactory[T],
]:
    ...


def context_factory(
    provider_class: typing.Union[
        typing.Type[_datastructures.AsyncProvider[T]],
        typing.Type[_datastructures.Provider[T]],
    ],
    context_class: typing.Union[
        typing.Type[_datastructures.AbstractAsyncContext[T]],
        typing.Type[_datastructures.AbstractSyncContext[T]],
    ],
    context_state_name: typing.Optional[str] = None,
) -> typing.Union[
    _datastructures.AbstractAsyncContextFactory[T],
    _datastructures.AbstractSyncContextFactory[T],
]:
    return _ContextFactory(
        provider_class, context_class, context_state_name
    )  # type:ignore


generate_state_name = _ContextFactory.generate_state_name



__version__ = "1.1.0"
__version_info__ = tuple(
    map(
        lambda val: int(val) if val.isnumeric() else val,
        __version__.split("."),
    )
)

from . import exc
from ._datastructures import (
    ContextGetter,
    ImmutableAsyncProvider,
    ImmutableSyncProvider,
    StateWrapper,
)
from .context import AsyncContext, SyncContext
from . import ensure_context
from .factory import context_factory

__all__ = [
    "AsyncContext",
    "SyncContext",
    "context_factory",
    "exc",
    "ensure_context",
    "ContextGetter",
    "ImmutableSyncProvider",
    "ImmutableAsyncProvider",
    "StateWrapper",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['context_handler', 'context_handler.adapters']

package_data = \
{'': ['*']}

install_requires = \
['typing-extensions>=3.10.0,<4.0.0']

extras_require = \
{'fastapi': ['fastapi>=0.70.0,<0.71.0'], 'sanic': ['sanic>=21.9.1,<22.0.0']}

setup_kwargs = {
    'name': 'context-handler',
    'version': '1.1.0',
    'description': '',
    'long_description': None,
    'author': 'Gustavo Correa',
    'author_email': 'self.gustavocorrea@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



class ContextError(Exception):
    """Base class for all Context errors"""


class ContextNotInitializedError(ContextError):
    """Raised when trying to get unitialized context"""


class ClosedContext(ContextError):
    """Raised when trying to use closed context"""

    def __init__(self, exc: Exception, *args: object) -> None:
        self.exc = exc
        super().__init__(*args)


class InvalidState(ContextError):
    """Raised when state is somehow invalid"""


class NoProviderInState(InvalidState):
    """State App does not have provider"""



import typing

import sanic as _sanic

from context_handler import _datastructures
from context_handler.factory import generate_state_name


def setup_context_cleaner_middleware(app: _sanic.Sanic):
    async def _context_cleaner_middleware(
        request: _sanic.Request,
        response: _sanic.HTTPResponse,
    ):
        app_state_dict = vars(request.app.ctx)
        request_state_dict = vars(request.ctx)
        provider_list = _get_provider_list(app_state_dict)
        contexts_from_provider = _get_contexts_from_providers(
            provider_list, request_state_dict
        )
        contexts_by_type = _get_contexts_by_type(request_state_dict)
        await _close_active_contexts(
            frozenset(contexts_from_provider + contexts_by_type)
        )
        return response

    app.on_response(_context_cleaner_middleware)


def _get_provider_list(
    _app_state_dict: typing.Dict[str, typing.Any]
) -> typing.List[
    typing.Union[_datastructures.Provider, _datastructures.AsyncProvider]
]:
    def _gen():
        for value in _app_state_dict:
            if isinstance(
                value,
                (_datastructures.Provider, _datastructures.AsyncProvider),
            ):
                yield value

    return list(_gen())


def _get_contexts_from_providers(
    provider_list: typing.List[
        typing.Union[_datastructures.Provider, _datastructures.AsyncProvider]
    ],
    request_state_dict: typing.Dict[str, typing.Any],
):
    def _gen():
        for provider in provider_list:
            if (
                context := request_state_dict.get(
                    generate_state_name(type(provider))
                )
            ) is not None:
                yield context

    return list(_gen())


def _get_contexts_by_type(request_state_dict: typing.Dict[str, typing.Any]):
    def _gen():
        for value in request_state_dict.values():
            if isinstance(
                value,
                (
                    _datastructures.AbstractSyncContext,
                    _datastructures.AbstractAsyncContext,
                ),
            ):
                yield value

    return list(_gen())


async def _close_active_contexts(
    contexts: typing.FrozenSet[
        typing.Union[
            _datastructures.AbstractAsyncContext,
            _datastructures.AbstractSyncContext,
        ]
    ]
):
    for ctx in contexts:
        if ctx.client is not None:
            provider = ctx.get_provider()
            if provider.is_closed(ctx.client):
                if isinstance(provider, _datastructures.AsyncProvider):
                    await provider.close_client(ctx.client)
                else:
                    provider.close_client(ctx.client)
